@@ -4,6 +4,7 @@ use v5.26;
 
 package ChordPro;
 
+use ChordPro::Files;
 use ChordPro::Utils;
 use ChordPro::Chords;
 use ChordPro::Output::Common;
@@ -74,7 +75,7 @@ our $config;
 package ChordPro;
 
 use ChordPro::Paths;
-use Encode qw(decode decode_utf8 encode_utf8);
+use Encode qw(decode_utf8);
 
 sub import {
     # Add private library.
@@ -88,6 +89,11 @@ sub import {
 sub ::run {
     binmode(STDERR, ':utf8');
     binmode(STDOUT, ':utf8');
+    for ( @ARGV ) {
+	next if ref ne "";
+	$_ = decode_utf8($_);
+    }
+
     $options = app_setup( "ChordPro", $VERSION );
     $options->{trace}   = 1 if $options->{debug};
     $options->{verbose} = 1 if $options->{trace};
@@ -108,6 +114,10 @@ sub chordpro {
     # Establish backend.
     my $of = $options->{output};
 
+    # Progress callback, if any.
+    progress( callback => $options->{progress_callback}, index => -1 )
+      if $options->{progress_callback};
+
     if ( $options->{'convert-config'} ) {
 	die("\"--convert-config\" requires a single config file name\n")
 	  if @ARGV;
@@ -189,8 +199,14 @@ sub chordpro {
 
     # Check for metadata in filelist. Actually, this works on the
     # command line as well, but don't tell anybody.
+    progress( phase => "Parsing", index => 0, total => 0+@ARGV )
+      if @ARGV > 1;
     foreach my $file ( @ARGV ) {
 	my $opts;
+	if ( $file =~ /^--((?:sub)?title)\s*(.*)/ ) {
+	    $options->{$1} = $2;
+	    next;
+	}
 	if ( $file =~ /(^|\s)--(?:meta|config|define)\b/ ) {
 	    # Break into words.
 	    my @w = Text::ParseWords::shellwords($file);
@@ -213,6 +229,7 @@ sub chordpro {
 	$opts->{generate} = $options->{generate};
 	# Wx runs on temp files, so pass real filename in.
 	$opts->{filesource} = $options->{filesource};
+	progress( msg => $file ) if @ARGV > 1;
 	$s->parse_file( $file, $opts );
     }
 
@@ -275,8 +292,7 @@ sub chordpro {
     # array of lines to be written.
     if ( $res && @$res > 0 ) {
         if ( $of && $of ne "-" ) {
-            open( my $fd, '>', $of );
-	    binmode( $fd, ":utf8" );
+            my $fd = fs_open( $of, '>:utf8' );
 	    push( @$res, '' ) unless $res->[-1] eq '';
 	    print { $fd } ( join( "\n", @$res ) );
 	    close($fd);
@@ -422,6 +438,17 @@ Requires the input to be strictly compliant to the ChordPro standard.
 
 This is enabled by default. See also B<--nostrict>.
 
+=item B<--subtitle=>I<XXX>
+
+Subtitle (for songbooks).
+
+=item B<--title=>I<XXX>
+
+Title (for songbooks).
+
+If specified and a table of contents is requested, a nice coverpage
+will be added.
+
 =item B<--toc> (short: B<-i>)
 
 Includes a table of contents.
@@ -632,6 +659,15 @@ the command line.
 This guarantees that the program is running with the default
 configuration.
 
+=item B<--print-template-config>
+
+Prints a simplified template configuration to standard output, and
+exits. The configuration is commented to explain its contents.
+
+The config contains most of the ChordPro configuration items, all
+commented out. It is easy to get started with configuring ChordPro
+by enabling and modifyng just a few items at a time.
+
 =item B<--print-default-config>
 
 Prints the default configuration to standard output, and exits.
@@ -643,8 +679,6 @@ The default configuration is commented to explain its contents.
 Prints the final configuration (after processing all system, user and
 other config files)  to standard output, and exits.
 
-=back
-
 =item B<--convert-config=>I<file>
 
 This option requires a single file name argument.
@@ -704,6 +738,7 @@ sub app_setup {
     my $defcfg = 0;		# handled locally
     my $fincfg = 0;		# handled locally
     my $deltacfg = 0;		# handled locally
+    my $tmplcfg = 0;		# handled locally
     my $dump_chords = 0;	# handled locally
 
     # Package name.
@@ -778,12 +813,18 @@ sub app_setup {
 	  "front-matter|cover=s",	# Front matter page(s)
 	  "back-matter=s",		# Back matter page(s)
 	  "filelist=s@",		# List of input files
+	  "title=s",			# Title (for books)
+	  "subtitle=s",			# Subtitle (for books)
 	  "meta=s\%",			# Command line meta data
 	  "decapo",			# remove capo
 	  "fragment|F",			# partial (incomplete) song
 	  "strict!",			# strict conformance
 
-          ### Standard Chordii Options ###
+	  ### Experimental ###
+
+	  "progress_callback=s",
+
+	  ### Standard Chordii Options ###
 
           "about|A+" => \$about,        # About...
           "chord-font|C=s",             # Sets chord font
@@ -827,9 +868,10 @@ sub app_setup {
           'nouserconfig|no-userconfig',
 	  'nodefaultconfigs|no-default-configs|X',
 	  'define=s%',
-	  'print-default-config' => \$defcfg,
-	  'print-final-config'   => \$fincfg,
-	  'print-delta-config'   => \$deltacfg,
+	  'print-default-config'  => \$defcfg,
+	  'print-final-config'    => \$fincfg,
+	  'print-delta-config'    => \$deltacfg,
+	  'print-template-config' => \$tmplcfg,
 	  'convert-config=s',
 
 	  # This aborts option scanning.
@@ -842,6 +884,7 @@ sub app_setup {
           'help-config'         => sub { $manual = 2 },
           'manual'              => \$manual,
           'verbose|v+',
+	  I			=> sub { $clo->{progress_callback} = "warn" },
           'trace',
           'debug+',
 
@@ -880,6 +923,7 @@ sub app_setup {
 
     $clo->{trace} ||= $clo->{debug};
     $clo->{verbose} ||= $clo->{trace};
+    $clo->{progress_callback} //= "warn" if $clo->{verbose};
 
     unless ( $ok ) {
         # GNU convention: message to STDERR upon failure.
@@ -924,10 +968,10 @@ sub app_setup {
                 foreach my $c ( @$_ ) {
 		    my $try = $c;
 		    # Check for resource names.
-		    if ( ! -r $try ) {
+		    if ( !fs_test( 'r', $try ) ) {
 			$try = CP->findcfg($c);
 		    }
-                    die("$c: $!\n") unless $try && -r $try;
+                    die("$c: $!\n") unless $try && fs_test( 'r', $try );
 		    $c = $try;
                 }
                 next;
@@ -936,7 +980,7 @@ sub app_setup {
 	    next if $clo->{nodefaultconfigs};
 	    next unless $configs{$config};
             $_ = [ $configs{$config} ];
-            undef($_) unless -r -f $_->[0];
+            undef($_) unless fs_test( 'fr', $_->[0] );
         }
     }
     # If no config was specified, and no default is available, force no.
@@ -945,30 +989,22 @@ sub app_setup {
     }
     $clo->{nosongconfig} ||= $clo->{nodefaultconfigs};
 
-    # Decode command line strings.
-    # File names are dealt with elsewhere.
-    for ( qw(transcode) ) {
-	next unless defined $clo->{$_};
-	$clo->{$_} = decode_utf8($clo->{$_});
-    }
-    ####TODO: Should decode all, and remove filename exception.
-    for ( keys %{ $clo->{define} } ) {
-	$clo->{define}->{$_} = decode_utf8($clo->{define}->{$_});
-    }
-
     # Plug in command-line options.
     @{$options}{keys %$clo} = values %$clo;
     $::options = $options;
     # warn(::dump($options), "\n") if $options->{debug};
 
+    if ( $tmplcfg ) {
+	use File::Copy;
+	my $cfg = File::Spec->catfile( CP->findresdirs("config")->[-1],
+				       "config.tmpl" );
+	binmode STDOUT => ':raw';
+	copy( $cfg, \*STDOUT );
+	exit 0;
+    }
     if ( $defcfg || $fincfg || $deltacfg ) {
-
-	if ( $defcfg ) {
-	    my $config = CP->findres( "chordpro.json", class => "config" );
-	    print loadlines( $config, { split => 0 } );
-	}
-	print ChordPro::Config::config_final($deltacfg)
-	  if $fincfg || $deltacfg;
+	print ChordPro::Config::config_final( default => $defcfg,
+					      delta   => $deltacfg );
 	exit 0;
     }
 
@@ -992,7 +1028,7 @@ sub app_setup {
 		next unless /\S/;
 		next if /^#/;
 		s/[\r\n]+$//;
-		push( @files, encode_utf8($_) );
+		push( @files, $_ );
 	    }
 	}
 	if ( @files ) {
@@ -1067,7 +1103,7 @@ use List::Util qw(uniq);
 
 sub ::runtimeinfo {
     my $level = shift // "normal";
-    my %i = %{runtime_info()};
+    my %i = %{runtime_info($level)};
     my $fmt0   = "  %-22.22s %-10s";
     my $fmt2   = $fmt0 . "\n";
     my $fmt3   = $fmt0 . " (%s)\n";
@@ -1158,6 +1194,7 @@ sub ::runtimeinfo {
 
 # Gather runtime details.
 sub runtime_info {
+    my $level = shift // "normal";
     my $cp = ChordPro::Paths->get;
     my $res;
 
@@ -1170,6 +1207,7 @@ sub runtime_info {
 	$_->{aux} = "Unsupported development snapshot"
 	  if $VERSION =~ /_/;
     }
+    return $res if $level eq "short";
 
     $res->{general}{perl} =
       { version => "$^V",
@@ -1229,7 +1267,7 @@ sub runtime_info {
 	no strict 'subs';
 	push( @p,
 	      { name => "wxPerl",    version => $dd->($Wx::VERSION)  },
-	      { name => "wxWidgets", version => $dd->(Wx::wxVERSION) } );
+	      { name => "wxWidgets", version => $dd->($Wx::wxVERSION) } );
     }
 
     local $SIG{__WARN__} = sub {};
@@ -1284,7 +1322,7 @@ sub runtime_info {
     $res->{metadata} = $::config->{metadata}{keys};
 
     @p = ( qw(title subtitle),
-	   ( grep { !/^(sub)?title$/ } sort(@{$::config->{metadata}{keys}}) ),
+	   ( grep { !/^(sub)?title$/ } sort(@{$::config->{metadata}{keys}//[]}) ),
 	   grep { !/^(sub)?title$/ } (keys(%{ChordPro::Song::_directives()})) );
     $res->{directives} = [ @p ];
 
