@@ -14,22 +14,27 @@ use feature qw( signatures state );
 no warnings "experimental::signatures";
 
 use ChordPro;
+use ChordPro::Files;
 use ChordPro::Paths;
 use ChordPro::Utils;
-use File::LoadLines;
 use File::Spec;
 use Scalar::Util qw(reftype);
 use List::Util qw(any);
 use Storable 'dclone';
 use Hash::Util;
+use Ref::Util qw( is_arrayref is_hashref );
 
 #sub hmerge($$;$);
 #sub clone($);
 #sub default_config();
 
+sub new ( $pkg, $cf = {} ) {
+    bless $cf => $pkg;
+}
+
 sub pristine_config {
     use ChordPro::Config::Data;
-    ChordPro::Config::Data::config();
+    __PACKAGE__->new(ChordPro::Config::Data::config());
 }
 
 sub configurator ( $opts = undef ) {
@@ -39,7 +44,7 @@ sub configurator ( $opts = undef ) {
     unless ( $opts ) {
         my $cfg = pristine_config();
         $config = $cfg;
-	config_split_fc_aliases($cfg);
+	$cfg->split_fc_aliases;
         $options = { verbose => 0 };
         process_config( $cfg, "<builtin>" );
         $cfg->{settings}->{lineinfo} = 0;
@@ -56,9 +61,6 @@ sub configurator ( $opts = undef ) {
     warn("Reading: <builtin>\n") if $verbose > 1;
     my $cfg = pristine_config();
 
-    # This is easier than splitting out manually :)
-    config_split_fc_aliases($cfg);
-
     # Default first.
     @cfg = prep_configs( $cfg, "<builtin>" );
     # Bubble default config to be the first.
@@ -68,7 +70,7 @@ sub configurator ( $opts = undef ) {
     my $add_config = sub {
         my $fn = shift;
         $cfg = get_config( $fn );
-        push( @cfg, prep_configs( $cfg, $fn ) );
+        push( @cfg, $cfg->prep_configs($fn) );
     };
 
     foreach my $c ( qw( sysconfig userconfig config ) ) {
@@ -88,7 +90,7 @@ sub configurator ( $opts = undef ) {
             splice( @cfg, $a, 1 );
             redo;
         }
-        print STDERR ("Config[$a]: ", $cfg[$a]->{_src}, "\n" )
+        warn("Config[$a]: ", $cfg[$a]->{_src}, "\n" )
           if $verbose;
     }
 
@@ -145,23 +147,26 @@ sub configurator ( $opts = undef ) {
     prpadd2cfg( $cfg, %{$options->{define}} );
 
     # Sanitize added extra entries.
-    for ( qw(title subtitle footer) ) {
-        delete($cfg->{pdf}->{formats}->{first}->{$_})
-          if ($cfg->{pdf}->{formats}->{first}->{$_} // 1) eq "";
-        for my $class ( qw(title first default) ) {
-            my $t = $cfg->{pdf}->{formats}->{$class}->{$_};
-            next unless $t;
-            die("Config error in pdf.formats.$class.$_: not an array\n")
-              unless ref($t) eq 'ARRAY';
-            if ( ref($t->[0]) ne 'ARRAY' ) {
-                $t = [ $t ];
-            }
-            my $tt = $_;
-            for ( @$t) {
-                die("Config error in pdf.formats.$class.$tt: ",
-                    scalar(@$_), " fields instead of 3\n")
-                  if @$_ && @$_ != 3;
-            }
+    for my $format ( qw(title subtitle footer) ) {
+        delete($cfg->{pdf}->{formats}->{first}->{$format})
+          if ($cfg->{pdf}->{formats}->{first}->{$format} // 1) eq "";
+        for my $c ( qw(title first default filler) ) {
+	    for my $class ( $c, $c."-even" ) {
+		my $t = $cfg->{pdf}->{formats}->{$class}->{$format};
+		# Allowed: null, false, [3], [[3], ...].
+		next unless defined $t;
+		$cfg->{pdf}->{formats}->{$class}->{$format} = ["","",""], next
+		  unless $t;
+		die("Config error in pdf.formats.$class.$format: not an array\n")
+		  unless is_arrayref($t);
+		$t = [ $t ] unless is_arrayref($t->[0]);
+		for ( @$t) {
+		    die("Config error in pdf.formats.$class.$format: ",
+			scalar(@$_), " fields instead of 3\n")
+		      if @$_ && @$_ != 3;
+		}
+		$cfg->{pdf}->{formats}->{$class}->{$format} = $t;
+	    }
         }
     }
 
@@ -263,22 +268,21 @@ sub get_config ( $file ) {
     $file = expand_tilde($file);
 
     if ( $file =~ /\.json$/i ) {
-        if ( open( my $fd, "<:raw", $file ) ) {
-            my $new = json_load( loadlines( $fd, { split => 0 } ), $file );
+        if ( my $lines = fs_load( $file, { split => 1, fail => "soft" } ) ) {
+            my $new = json_load( join( "\n", @$lines, '' ), $file );
             precheck( $new, $file );
-            close($fd);
-            return $new;
+            return __PACKAGE__->new($new);
         }
         else {
             die("Cannot open config $file [$!]\n");
         }
     }
     elsif ( $file =~ /\.prp$/i ) {
-        if ( -e -f -r $file ) {
+        if ( fs_test( efr => $file ) ) {
             require ChordPro::Config::Properties;
-            my $cfg = new Data::Properties;
+            my $cfg = Data::Properties->new;
             $cfg->parse_file($file);
-            return $cfg->data;
+            return __PACKAGE__->new($cfg->data);
         }
         else {
             die("Cannot open config $file [$!]\n");
@@ -309,10 +313,12 @@ sub prep_configs ( $cfg, $src ) {
         }
         my $cfg = get_config($c);
         # Recurse.
-        push( @res, prep_configs( $cfg, $c ) );
+        push( @res, $cfg->prep_configs($c) );
     }
 
     # Push this and return.
+    $cfg->split_fc_aliases;
+    $cfg->expand_font_shortcuts;
     push( @res, $cfg );
     return @res;
 }
@@ -353,12 +359,13 @@ sub process_config ( $cfg, $file ) {
         $cfg->{_chords} = delete $cfg->{chords};
         ChordPro::Chords::pop_parser();
     }
-    config_split_fc_aliases($cfg);
-    config_expand_font_shortcuts($cfg);
+    $cfg->split_fc_aliases;
+    $cfg->expand_font_shortcuts;
 }
 
 # Expand pdf.fonts.foo: bar to pdf.fonts.foo { description: bar }.
-sub config_expand_font_shortcuts ( $cfg ) {
+
+sub expand_font_shortcuts ( $cfg ) {
     return unless exists $cfg->{pdf}->{fonts};
     for my $f ( keys %{$cfg->{pdf}->{fonts}} ) {
 	next if ref($cfg->{pdf}->{fonts}->{$f}) eq 'HASH';
@@ -380,14 +387,20 @@ sub config_expand_font_shortcuts ( $cfg ) {
 	    }
 	    else {
 		$i->{description} = $v;
+		$i->{description} .= " " . delete($i->{size})
+		  if $i->{size};
 	    }
 	    $_ = $i;
 	}
     }
 }
 
-sub config_split_fc_aliases ( $cfg ) {
-    # Split fontconfig aliases into individual entries.
+use Storable qw(dclone);
+
+# Split fontconfig aliases into separate entries.
+
+sub split_fc_aliases ( $cfg ) {
+
     if ( $cfg->{pdf}->{fontconfig} ) {
 	# Orig.
 	my $fc = $cfg->{pdf}->{fontconfig};
@@ -406,44 +419,113 @@ sub config_split_fc_aliases ( $cfg ) {
     }
 }
 
-sub config_final ( $delta ) {
-    $options->{'cfg-print'} = 1;
-    my $cfg = configurator($options);
+# Reverse of config_expand_font_shortcuts.
 
-    if ( $delta ) {
-        my $def = pristine_config();
-	config_split_fc_aliases($def);
-        $cfg->reduce($def);
+sub simplify_fonts( $cfg ) {
+
+    return $cfg unless $cfg->{pdf}->{fonts};
+
+    foreach my $font ( keys %{$cfg->{pdf}->{fonts}} ) {
+	for ( $cfg->{pdf}->{fonts}->{$font} ) {
+	    next unless is_hashref($_);
+
+	    delete $_->{color}
+	      if $_->{color} && $_->{color} eq "foreground";
+	    delete $_->{background}
+	      if $_->{background} && $_->{background} eq "background";
+
+	    if ( exists( $_->{file} ) ) {
+		delete $_->{description};
+		delete $_->{name};
+	    }
+	    elsif ( exists( $_->{description} ) ) {
+		delete $_->{name};
+		if ( $_->{size} && $_->{description} !~ /\s+[\d.]+$/ ) {
+		    $_->{description} .= " " . $_->{size};
+		}
+		delete $_->{size};
+		$_ = $_->{description} if keys %$_ == 1;
+	    }
+	    elsif ( exists( $_->{name} )
+		    && exists( $_->{size})
+		    && keys %$_ == 2
+		  ) {
+		$_ = $_->{name} .= " " . $_->{size};
+	    }
+	}
     }
+}
+
+sub config_final ( %args ) {
+    my $delta   = $args{delta} || 0;
+    my $default = $args{default} || 0;
+    $options->{'cfg-print'} = 1;
+
+    my $defcfg;			# pristine config
+    my $cfg;			# actual config
+    if ( $default || $delta ) {
+	local $options->{nosysconfig} = 1;
+	local $options->{nouserconfig} = 1;
+	local $options->{noconfig} = 1;
+	$defcfg = pristine_config();
+	split_fc_aliases($defcfg);
+	expand_font_shortcuts($defcfg);
+	if ( $delta ) {
+	    delete $defcfg->{chords};
+	    delete $defcfg->{include};
+	}
+	bless $defcfg => __PACKAGE__;
+	$cfg = $defcfg if $default;
+    }
+
+    $cfg //= configurator($options);
+
+    # Remove unwanted data.
     $cfg->unlock;
     $cfg->{tuning} = delete $cfg->{_tuning};
-    delete($cfg->{tuning}) if $delta && !defined($cfg->{tuning});
-    $cfg->{chords} = delete $cfg->{_chords};
+    if ( $delta ) {
+	for ( qw( tuning ) ) {
+	    delete($cfg->{$_}) unless defined($cfg->{$_});
+	}
+	for my $f ( keys( %{$cfg->{pdf}{fonts}} ) ) {
+	    for ( qw( background color ) ) {
+		next if defined($defcfg->{pdf}{fonts}{$f}{$_});
+		delete($cfg->{pdf}{fonts}{$f}{$_});
+		delete($defcfg->{pdf}{fonts}{$f}{$_});
+	    }
+	}
+    }
+    delete $cfg->{_chords};
     delete $cfg->{chords};
     delete $cfg->{_src};
-    $cfg->lock;
 
-    if ( $ENV{CHORDPRO_CFGPROPS} ) {
-        cfg2props($cfg);
-    }
-    else {
-	my $parser = JSON::Relaxed::Parser->new( key_order => 1 );
+    my $parser = JSON::Relaxed::Parser->new( key_order => 1 );
 
+    # Load schema.
+    my $schema = do {
 	my $schema = CP->findres( "config.schema", class => "config" );
-	my $data = loadlines( $schema, { split => 0 } );
-	$data = $parser->decode($data);
-	$schema = $data;
+	my $data = fs_load( $schema, { split => 0 } );
+	$parser->decode($data);
+    };
 
-	my $config = CP->findres( "chordpro.json", class => "config" );
-	$data = loadlines( $config, { split => 0 } );
-	my $new = $parser->decode($data);
-
-	$cfg->unlock;
-	my $tmp = hmerge($new,$cfg);
-	my $res = $parser->encode( data => $tmp, pretty => 1, schema => $schema );
-	$cfg->lock;
-	$res;
+    # Delta cannot handle reference config yet.
+    if ( $delta ) {
+	$defcfg->unlock;
+	$cfg->reduce( $defcfg );
+	return $parser->encode( data => {%$cfg},
+				pretty => 1, schema => $schema );
     }
+
+    my $config = do {
+	my $config = CP->findres( "chordpro.json", class => "config" );
+	my $data = fs_load( $config, { split => 0 } );
+	$parser->decode($data);
+    };
+
+    #    $cfg = hmerge( $config, $cfg );
+    $cfg->simplify_fonts;
+    return $parser->encode( data => {%{$cfg}},
+			    pretty => 1, schema => $schema );
 }
 
 sub convert_config ( $from, $to ) {
@@ -455,19 +537,20 @@ sub convert_config ( $from, $to ) {
     # First find and process the schema.
     my $schema = CP->findres( "config.schema", class => "config" );
     my $o = { split => 0, fail => 'soft' };
-    my $data = loadlines( $schema, $o );
+    my $data = fs_load( $schema, $o );
     die("$schema: ", $o->{error}, "\n") if $o->{error};
     $schema = $parser->decode($data);
 
     # Then load the config to be converted.
     my $new;
-    $o = { split => 0, fail => 'soft' };
-    $data = loadlines( $from, $o );
+    $o = { split => 1, fail => 'soft' };
+    $data = fs_load( $from, $o );
     die("Cannot open config $from [", $o->{error}, "]\n") if $o->{error};
+    $data = join( "\n", @$data );
 
     if ( $data =~ /^\s*#/m ) {	# #-comments -> prp
 	require ChordPro::Config::Properties;
-	my $cfg = new Data::Properties;
+	my $cfg = Data::Properties->new;
 	$cfg->parse_file($from);
 	$new = $cfg->data;
     }
@@ -478,7 +561,7 @@ sub convert_config ( $from, $to ) {
     # And re-encode it using the schema.
     my $res = $parser->encode( data => $new, pretty => 1,
 			       nounicodeescapes => 1, schema => $schema );
-
+use DDP; p $res;
     # Add trailer.
     $res .= "\n// End of Config.\n";
 
@@ -504,13 +587,13 @@ sub cfg2props ( $o, $path = "" ) {
     if ( !defined $o ) {
         $ret .= "$path: undef\n";
     }
-    elsif ( UNIVERSAL::isa( $o, 'HASH' ) ) {
+    elsif ( is_hashref($o) ) {
         $path .= "." unless $path eq "";
         for ( sort keys %$o ) {
             $ret .= cfg2props( $o->{$_}, $path . $_  );
         }
     }
-    elsif ( UNIVERSAL::isa( $o, 'ARRAY' ) ) {
+    elsif ( is_arrayref($o) ) {
         $path .= "." unless $path eq "";
         for ( my $i = 0; $i < @$o; $i++ ) {
             $ret .= cfg2props( $o->[$i], $path . "$i" );
@@ -568,6 +651,7 @@ sub _augment ( $self, $hash, $path ) {
         warn("Config augment error: unknown item $path$key\n")
           unless exists $self->{$key}
             || $path =~ /^pdf\.(?:info|fonts|fontconfig)\./
+            || $path =~ /^pdf\.formats\.\w+-even\./
             || $path =~ /^meta\./
             || $key =~ /^_/;
 
@@ -637,13 +721,13 @@ sub reduce ( $self, $hash ) {
 
     my $locked = $self->is_locked;
 
-    warn("O: ", qd($hash,1), "\n") if DEBUG;
-    warn("N: ", qd($self,1), "\n") if DEBUG;
+    warn("O: ", qd($hash,1), "\n") if DEBUG > 1;
+    warn("N: ", qd($self,1), "\n") if DEBUG > 1;
     my $state = _reduce( $self, $hash, "" );
 
     $self->lock if $locked;
 
-    warn("== ", qd($self,1), "\n") if DEBUG;
+    warn("== ", qd($self,1), "\n") if DEBUG > 1;
     return $self;
 }
 
@@ -665,9 +749,10 @@ sub _reduce ( $self, $orig, $path ) {
 
             warn("Config reduce error: unknown item $path$key\n")
               unless exists $self->{$key}
-                || $key =~ /^_/;
+                || $key =~ /^_/
+                || $path =~ /^pdf\/\.fonts\./;
 
-            unless ( defined $orig->{$key} ) {
+            unless ( exists $orig->{$key} ) {
                 warn("D: $path$key\n") if DEBUG;
                 delete $self->{$key};
                 $state //= 'M';
@@ -802,6 +887,8 @@ sub hmerge( $left, $right, $path = "" ) {
           unless exists $res{$key}
             || $path eq "pdf.fontconfig."
             || $path =~ /^pdf\.(?:info|fonts)\./
+            || $path =~ /^pdf\.formats\.\w+-even\./
+            || ( $path =~ /^pdf\.formats\./ && $key =~ /\w+-even$/ )
             || $path =~ /^meta\./
             || $path =~ /^delegates\./
             || $path =~ /^parser\.preprocess\./
@@ -881,17 +968,13 @@ sub precheck ( $cfg, $file ) {
     $p = sub {
         my ( $o, $path ) = @_;
         $path //= "";
-        if ( !defined $o ) {
-            warn("$file: Undefined config \"$path\" (using 'false')\n");
-            $_[0] = '';
-        }
-        elsif ( UNIVERSAL::isa( $o, 'HASH' ) ) {
+        if ( is_hashref($o) ) {
             $path .= "." unless $path eq "";
             for ( sort keys %$o ) {
                 $p->( $o->{$_}, $path . $_  );
             }
         }
-        elsif ( UNIVERSAL::isa( $o, 'ARRAY' ) ) {
+        elsif ( is_arrayref($o) ) {
             $path .= "." unless $path eq "";
             for ( my $i = 0; $i < @$o; $i++ ) {
                 $p->( $o->[$i], $path . "$i" );
