@@ -20,8 +20,8 @@ use File::Spec;
 use File::Temp ();
 use File::LoadLines;
 use feature 'state';
-use Encode 'decode_utf8';
 
+use ChordPro::Files;
 use ChordPro::Paths;
 use ChordPro::Utils;
 use Text::ParseWords qw(shellwords);
@@ -58,7 +58,7 @@ sub abc2svg( $song, %args ) {
 	if ( ($config->{delegates}{abc}{config} // "default") ne "default" ) {
 	    warn("ABC: delegates.abc.config is no longer used.\n");
 	    warn("ABC: Config \"default.abc\" will be loaded instead.\n")
-	      if !$abc2svg->{external} && -s "default.abc";
+	      if !$abc2svg->{external} && fs_test( s => "default.abc" );
 	}
     }
 
@@ -76,7 +76,7 @@ sub abc2svg( $song, %args ) {
 
     # External tools usually process a default.abc.
     warn("ABC: Using config \"default.abc\".\n")
-      if index( $abc2svg->{method}, QUICKJSXS ) < 0 && -s "default.abc";
+      if index( $abc2svg->{method}, QUICKJSXS ) < 0 && fs_test( s => "default.abc" );
 
     my $prep = make_preprocessor( $cfg->{preprocess} );
 
@@ -131,7 +131,7 @@ sub abc2svg( $song, %args ) {
 	@pre = ();
     }
     my $kv = {};
-    $kv = parse_kv( @pre ) if @pre;
+    $kv = parse_kvm( @pre ) if @pre;
     $kv = { %$kv, %{$elt->{opts}} };
     $kv->{split} //= 1;		# less overhead. really.
     $kv->{scale} ||= 1;		# with id: design scale
@@ -147,7 +147,7 @@ sub abc2svg( $song, %args ) {
 
     # Create the temp file for the ABC source.
     my $fd;
-    unless ( open( $fd, '>:utf8', $src ) ) {
+    unless ( $fd = fs_open( $src, '>:utf8' ) ) {
 	warn("Error in ABC embedding: $src: $!\n");
 	return;
     }
@@ -191,7 +191,7 @@ sub abc2svg( $song, %args ) {
 	   },
 	   loadjs    => sub {
 	       my ( $fn, $relay, $onerror ) = @_;
-	       if ( -s -r "$base/$fn" ) {
+	       if ( fs_test( sr => "$base/$fn" ) ) {
 		   $js->eval(slurp("$base/$_[0]"));
 		   $relay->() if $relay;
 	       }
@@ -213,7 +213,7 @@ sub abc2svg( $song, %args ) {
 
 	warn( "+ QuickJS_XS[", CP->display($base), "] $src\n") if DEBUG;
 	my $hooks = "$base/../hooks.js";
-	undef $hooks unless -s $hooks;
+	undef $hooks unless fs_test( s => $hooks );
 
 	eval {
 	    $js->eval( slurp("$base/abc2svg-1.js") );
@@ -237,7 +237,7 @@ sub abc2svg( $song, %args ) {
 	undef $js;
 
 	if ( DEBUG ) {
-	    open( my $fd, '>:utf8', $out );
+	    my $fd = fs_open( $out, '>:utf8' );
 	    print $fd join("\n", @lines), "\n";
 	    close($fd);
 	}
@@ -359,7 +359,7 @@ sub abc2svg( $song, %args ) {
 	open( $fd, '>:utf8', $svg );
 	print( $fd $_, "\n" ) for @data;
 	close($fd);
-	warn("SVG: ", 1+$lines, " lines (", -s $svg, " bytes)\n") if DEBUG > 1;
+	warn("SVG: ", 1+$lines, " lines (", fs_test( s => $svg ), " bytes)\n") if DEBUG > 1;
     }
 
     my $scale;
@@ -407,7 +407,7 @@ sub slurp {
 sub info {
     state $info = { handler => "" };
     my $ctl = $::config->{delegates}->{abc};
-    my $handler = $ctl->{handler};
+    my $handler = $ctl->{handler} // "abc2svg";
 
     # Use cached info, but allow handler change between songs.
     return $info if $handler eq $info->{handler};
@@ -439,7 +439,8 @@ sub info {
 	if ( $handler ne "quickjs_qjs" && have_xs() ) {
 	    $info->{method} = QUICKJSXS;
 	}
-	elsif ( $handler ne "quickjs_xs" && ($exe = CP->findexe("qjs")) ) {
+	elsif ( $handler ne "quickjs_xs"
+		&& ($exe = CP->findexe("qjs", silent => 1 )) ) {
 	    $info->{method} = QUICKJS;
 	}
 	if ( $info->{method} ) {
@@ -460,7 +461,7 @@ sub info {
     }
 
     elsif ( $handler eq "abc2svg" && $ctl->{program}
-	    && ( $exe = CP->findexe($ctl->{program}) ) ) {
+	    && ( $exe = CP->findexe($ctl->{program}, silent => 1 ) ) ) {
 	$info->{handler} = $handler;
 	$info->{method} = CP->display($exe);
 	$info->{info} = $info->{method};
@@ -481,7 +482,7 @@ sub options( $data ) {
     }
     @pre = () if @pre && !@data;	# no data found
     my $kv = {};
-    $kv = parse_kv( @pre ) if @pre;
+    $kv = parse_kvm( @pre ) if @pre;
     $kv->{align} //= ($kv->{center}//0) ? "center" : "left";
     $kv;
 }
