@@ -8,16 +8,16 @@ package ChordPro::Output::PDF::Writer;
 
 use strict;
 use warnings;
-use Encode;
 use Text::Layout;
 use IO::String;
 use Carp;
 use utf8;
 
+use ChordPro::Files;
 use ChordPro::Paths;
 use ChordPro::Utils qw( expand_tilde demarkup min is_corefont );
 use ChordPro::Output::Common qw( fmt_subst prep_outlines );
-use File::LoadLines qw(loadlines);
+use Ref::Util qw( is_hashref );
 
 # For regression testing, run perl with PERL_HASH_SEED set to zero.
 # This eliminates the arbitrary order of font definitions and triggers
@@ -413,7 +413,7 @@ sub add_object {
     my $w = $o->width  * $scale_x;
     my $h = $o->height * $scale_y;
 
-    warn( sprintf("add_object x=%.1f y=%.1f w=%.1f h=%.1f scale=%.1f,%.1f) %s\n",
+    warn( sprintf("add_object x=%.1f y=%.1f w=%.1f h=%.1f scale=%.1f,%.1f %s\n",
 		  $x, $y, $w, $h, $scale_x, $scale_y, $ha,
 		 ) ) if $config->{debug}->{images};
 
@@ -687,7 +687,7 @@ sub init_fonts {
     for my $fontdir ( @d ) {
 	next unless $fontdir;
 	$fontdir = expand_tilde($fontdir);
-	if ( -d $fontdir ) {
+	if ( fs_test( d => $fontdir ) ) {
 	    $self->{pdfapi}->can("addFontDirs")->($fontdir);
 	    $fc->add_fontdirs($fontdir);
 	    push( @dirs, $fontdir );
@@ -724,7 +724,7 @@ sub init_fonts {
 			       FreeMonoBold.ttf
 			       FreeMonoOblique.ttf
 			    ) ) {
-		$have = 0, last unless -f -s "$dir/$font";;
+		$have = 0, last unless fs_test( fs => "$dir/$font" );
 	    }
 	    $remap = "free", last if $have;
 	}
@@ -736,7 +736,7 @@ sub init_fonts {
 	my @fam = split( /\s*,\s*/, $ff );
 	foreach my $s ( keys( %{ $ps->{fontconfig}->{$ff} } ) ) {
 	    my $v = $ps->{fontconfig}->{$ff}->{$s};
-	    if ( UNIVERSAL::isa( $v, 'HASH' ) ) {
+	    if ( is_hashref($v) ) {
 		my $file = delete( $v->{file} );
 		$fc->register_font( $file, $fam[0], $s, $v );
 	    }
@@ -785,7 +785,8 @@ sub init_pangofont {
 	$font->{_ff} = $ff;
 	$font->{fd}->set_shaping( $font->{fd}->get_shaping || $font->{shaping}//0);
 	$font->{size} = $font->{fd}->get_size if $font->{fd}->get_size;
-    };
+	1;
+    } or return;
     $font->{fd};
 }
 
@@ -830,8 +831,7 @@ sub show_vpos {
 
 sub embed {
     my ( $self, $file ) = @_;
-    $file = encode_utf8($file);
-    return unless -f $file;
+    return unless fs_test( 'f', $file );
 
     # Borrow some routines from PDF Api.
     *PDFNum = \&{$self->{pdfapi} . '::Basic::PDF::Utils::PDFNum'};
@@ -841,14 +841,14 @@ sub embed {
     # Apparently the 'hidden' flag does not hide it completely,
     # so give it a rect outside the page.
     my $a = $self->{pdfpage}->annotation();
-    $a->text( loadlines( $file, { split => 0 } ),
+    $a->text( fs_load( $file, { fail => "soft", split => 0 } ),
 	      -open => 0, -rect => [0,0,-1,-1] );
     $a->{T} = PDFStr("ChordProSong");
     $a->{F} = PDFNum(2);		# hidden
 
     # The config.
     $a = $self->{pdfpage}->annotation();
-    $a->text( ChordPro::Config::config_final(0),
+    $a->text( ChordPro::Config::config_final(),
 	      -open => 0, -rect => [0,0,-1,-1]);
     $a->{T} = PDFStr("ChordProConfig");
     $a->{F} = PDFNum(2);		# hidden
