@@ -9,14 +9,21 @@ class ChordPro::Paths;
 
 my $instance;
 
-method get :common ( $reset = 0 ) {
+# Work around Object::Pad 0.817 breakage.
+#method get :common ( $reset = 0 ) {
+#    undef $instance if $reset;
+#    $instance //= $class->new;
+#}
+
+sub get( $class, $reset = 0 ) {
     undef $instance if $reset;
     $instance //= $class->new;
 }
 
 use Cwd qw(realpath);
-use File::Spec::Functions qw( catfile catdir splitpath catpath );
+use File::Spec::Functions qw( catfile catdir splitpath catpath file_name_is_absolute );
 use File::HomeDir;
+use ChordPro::Files;
 
 field $home      :reader;	# dir
 field $configdir :reader;	# dir
@@ -34,7 +41,7 @@ BUILD {
     my $app = "ChordPro";
     my $app_lc = lc($app);
 
-    $pathsep = $self->is_msw ? ';' : ':';
+    $pathsep = is_msw ? ';' : ':';
 
     $home     = realpath( $ENV{HOME} = File::HomeDir->my_home );
 
@@ -76,14 +83,14 @@ BUILD {
     }
 
     for ( @try ) {
-	next unless $_ && -d $_;
+	next unless $_ && fs_test( d => $_);
 	my $path = $self->normalize($_);
 	warn("Paths: configdir try $_ => $path\n") if $self->debug > 1;
-	next unless $path && -d $path;
+	next unless $path && fs_test( d => $path);
 	$configdir = $path;
 	for ( $self->normalize( catfile( $path, "$app_lc.prp" ) ),
 	      $self->normalize( catfile( $path, "$app_lc.json" ) ) ) {
-	    next unless $_ && -f $_;
+	    next unless $_ && fs_test( f => $_ );
 	    $configs->{userconfig} = $_;
 	    last;
 	}
@@ -93,7 +100,7 @@ BUILD {
 
     for ( $self->normalize(".$app_lc.json"),
 	  $self->normalize("$app_lc.json") ) {
-	    next unless $_ && -f $_;
+	    next unless $_ && fs_test( f => $_ );
 	$configs->{config} = $_;
 	last;
     }
@@ -108,8 +115,23 @@ BUILD {
     $privlib = $INC{'ChordPro.pm'} =~ s/\.pm$/\/lib/r;
 
     # Now for the resources.
+    $self->setup_resdirs;
+
+    # Check for packaged image.
+    for ( qw( Docker AppImage PPL ) ) {
+	next unless exists $ENV{uc($_)."_PACKAGED"}
+	  && $ENV{uc($_)."_PACKAGED"};
+	$packager = $_;
+	last;
+    }
+
+};
+
+# We need this to be able to re-establish the resdirs, e.g. after a change
+# of CHORDPRO_LIB.
+method setup_resdirs {
     $resdirs = [];
-    @try = ();
+    my @try = ();
     push( @try, $self->path($ENV{CHORDPRO_LIB}) )
       if defined($ENV{CHORDPRO_LIB});
     push( @try, $configdir ) if $configdir;
@@ -119,7 +141,7 @@ BUILD {
 	next unless $_;
 	my $path = $self->normalize($_);
 	warn("Paths: resdirs try $_ => $path\n") if $self->debug > 1;
-	next unless $path && -d $path;
+	next unless $path && fs_test( d => $path );
 	push( @$resdirs, $path );
     }
 
@@ -132,16 +154,7 @@ BUILD {
     unless ( @$resdirs ) {
 	warn("Paths: Cannot find resources, prepare for disaster\n");
     }
-
-    # Check for packaged image.
-    for ( qw( Docker AppImage PPL ) ) {
-	next unless exists $ENV{uc($_)."_PACKAGED"}
-	  && $ENV{uc($_)."_PACKAGED"};
-	$packager = $_;
-	last;
-    }
-
-};
+}
 
 method debug {
     # We need to take an env var into account, since the Paths
@@ -149,10 +162,6 @@ method debug {
     $ENV{CHORDPRO_DEBUG_PATHS} || $::config->{debug}->{paths} || 0;
 }
 
-method is_msw {
-    $^O =~ /mswin/i;
-}
-
 # Is absolute.
 
 method is_absolute ( $p ) {
@@ -189,7 +198,7 @@ method path ( $p = undef ) {
 	local $ENV{PATH} = $p;
 	my @p = File::Spec->path();
 	# On MSWindows, '.' is always prepended.
-	shift(@p) if $self->is_msw;
+	shift(@p) if is_msw;
 	return @p;
     }
     return File::Spec->path();
@@ -210,15 +219,28 @@ method pathcombine( @d ) {
 method findexe ( $p, %opts ) {
     my $try = $p;
     my $found;
-    if ( $self->is_msw ) {
+    if ( is_msw ) {
 	$try .= ".exe";
     }
+
+    if ( File::Spec->file_name_is_absolute($p)
+	 && ChordPro::Files::fs_test( fx => $p ) ) {
+	warn("Paths: findexe $p => ", $self->display($p), "\n")
+	  if $self->debug;
+	return $p;
+    }
+
     for ( $self->path ) {
 	my $e = catfile( $_, $try );
-	$found = realpath($e), last if -f -x $e;
+	$found = realpath($e), last if fs_test( fx => $e );
+    }
+    if ( $self->debug ) {
+	warn("Paths: findexe $p => ", $self->display($found), "\n");
+    }
+    elsif ( !$opts{silent} ) {
+	warn("Could not find $p in ",
+	     join( " ", map { qq{"$_"} } $self->path ), "\n");
     }
-    warn("Paths: findexe $p => ", $self->display($found), "\n")
-      if $self->debug;
     return $found;
 }
 
@@ -228,7 +250,7 @@ method findcfg ( $p ) {
     my $found;
     my @p;
     if ( $p =~ /\.\w+$/ ) {
-	$found = realpath($p) if -f -s $p;
+	$found = realpath($p) if fs_test( fs => $p );
 	@p = ( $p );
     }
     else {
@@ -239,7 +261,7 @@ method findcfg ( $p ) {
 	for ( @$resdirs ) {
 	    for my $cfg ( @p ) {
 		my $f = catfile( $_, "config", $cfg );
-		$found = realpath($f), last if -f -s $f;
+		$found = realpath($f), last if fs_test( fs => $f );
 	    }
 	}
     }
@@ -253,12 +275,17 @@ method findcfg ( $p ) {
 method findres ( $p, %opts ) {
     my $try = $p;
     my $found;
-    if ( defined $opts{class} ) {
-	$try = catfile( $opts{class}, $try );
+    if ( file_name_is_absolute($p) ) {
+	$found = realpath($p);
     }
-    for ( @$resdirs ) {
-	my $f = catfile( $_, $try );
-	$found = realpath($f), last if -f -s $f;
+    else {
+	if ( defined $opts{class} ) {
+	    $try = catfile( $opts{class}, $try );
+	}
+	for ( @$resdirs ) {
+	    my $f = catfile( $_, $try );
+	    $found = realpath($f), last if fs_test( fs => $f );
+	}
     }
     warn("Paths: findres", $opts{class} ? " [$opts{class}]" : "",
 	 " $p => ", $self->display($found), "\n")
@@ -276,7 +303,7 @@ method findresdirs ( $p, %opts ) {
     }
     for ( @$resdirs ) {
 	my $d = catdir( $_, $p );
-	push( @found, realpath($d) ) if -d $d;
+	push( @found, realpath($d) ) if fs_test( d => $d );
     }
     if ( $self->debug ) {
 	my $i = 0;
@@ -321,7 +348,7 @@ method sibling ( $orig, %opts ) {
 method siblingres ( $orig, $name, %opts ) {
     return unless defined $orig;
     my $try = $self->sibling( $orig, name => $name );
-    my $found = ( $try && -s $try )
+    my $found = ( $try && fs_test( s => $try ) )
       ? $try
       : $self->findres( $name, class => $opts{class} );
     return $found;
