@@ -13,6 +13,7 @@ use strict;
 use warnings;
 
 use ChordPro;
+use ChordPro::Files;
 use ChordPro::Paths;
 use ChordPro::Chords;
 use ChordPro::Chords::Appearance;
@@ -48,6 +49,7 @@ my $chorus_xpose = 0;
 my $chorus_xpose_dir = 0;
 
 # Memorized chords.
+my $cctag;			# current cc name
 my %memchords;			# all sections
 my $memchords;			# current section
 my $memcrdinx;			# chords tally
@@ -56,7 +58,8 @@ my $memorizing;			# if memorizing (a.o.t. recalling)
 # Keep track of unknown chords, to avoid dup warnings.
 my %warned_chords;
 
-my $re_chords;			# for chords
+our $re_chords;			# for chords
+my $propitems_re = propitems_re();
 my $intervals;			# number of note intervals
 my @labels;			# labels used
 
@@ -95,7 +98,7 @@ sub new {
     $xcmov = undef;
     upd_config();
 
-    $diag->{format} = $config->{diagnostics}->{format};
+    $diag->{format} = $opts->{diagformat} // $config->{diagnostics}->{format};
     $diag->{file}   = $filesource;
     $diag->{line}   = 0;
     $diag->{orig}   = "(at start of song)";
@@ -139,11 +142,10 @@ sub parse_song {
 	    }
 	}
 	if ( $cf ) {
-	    my $pp = JSON::PP->new->relaxed;
-	    my $precfg = $pp->decode($cf);
 	    my $prename = "__PRECFG__";
-	    ChordPro::Config::precheck( $precfg, $prename );
-	    push( @configs, ChordPro::Config::prep_configs( $precfg, $prename) );
+	    my $precfg = ChordPro::Config->new( json_load( $cf, $prename ) );
+	    $precfg->precheck($prename);
+	    push( @configs, $precfg->prep_configs($prename) );
 	}
     }
     # Load song-specific config, if any.
@@ -161,16 +163,16 @@ sub parse_song {
 	    warn("Config[song]: $cf\n") if $options->{verbose};
 	    my $have = ChordPro::Config::get_config( CP->findcfg($cf) );
 	    die("Missing config: $cf\n") unless $have;
-	    push( @configs, ChordPro::Config::prep_configs( $have, $cf) );
+	    push( @configs, $have->prep_configs($cf) );
 	}
 	else {
 	    for ( "prp", "json" ) {
 		( my $cf = $diag->{file} ) =~ s/\.\w+$/.$_/;
 		$cf .= ".$_" if $cf eq $diag->{file};
-		next unless -s $cf;
+		next unless fs_test( s => $cf );
 		warn("Config[song]: $cf\n") if $options->{verbose};
 		my $have = ChordPro::Config::get_config($cf);
-		push( @configs, ChordPro::Config::prep_configs( $have, $cf) );
+		push( @configs, $have->prep_configs($cf) );
 		last;
 	    }
 	}
@@ -217,9 +219,7 @@ sub parse_song {
     $config->unlock;
 
     if ( %$defs ) {
-	my $c = $config->hmerge( prp2cfg( $defs, $config ) );
-	bless $c => ref($config);
-	$config = $c;
+	prpadd2cfg( $config, %$defs );
     }
 
     for ( qw( transpose transcode decapo lyrics-only ) ) {
@@ -331,10 +331,15 @@ sub parse_song {
 	    $_ .= $cont;
 	}
 
+	# Uncomment this to allow \uDXXX\uDYYY (surrogate) escapes.
+	s/ \\u(d[89ab][[:xdigit:]]{2})\\u(d[cdef][[:xdigit:]]{2})
+	 / pack('U*', 0x10000 + (hex($1) - 0xD800) * 0x400 + (hex($2) - 0xDC00) )
+	   /igex;
+
 	# Uncomment this to allow \uXXXX escapes.
 	s/\\u([0-9a-f]{4})/chr(hex("0x$1"))/ige;
 	# Uncomment this to allow \u{XX...} escapes.
-	# s/\\u\{([0-9a-f]+)\}/chr(hex("0x$1"))/ige;
+	s/\\u\{([0-9a-f]+)\}/chr(hex("0x$1"))/ige;
 
 	$diag->{orig} = $_;
 	# Get rid of TABs.
@@ -346,9 +351,9 @@ sub parse_song {
 
 	for my $pp ( "all", "env-$in_context" ) {
 	    if ( $prep->{$pp} ) {
-		0&&warn("PRE:  ", $_, "\n");
+		$config->{debug}->{pp} && warn("PRE:  ", $_, "\n");
 		$prep->{$pp}->($_);
-		0&&warn("POST: ", $_, "\n");
+		$config->{debug}->{pp} && warn("POST: ", $_, "\n");
 		if ( /\n/ ) {
 		    my @a = split( /\n/, $_ );
 		    $_ = shift(@a);
@@ -376,10 +381,16 @@ sub parse_song {
 		last;
 	    }
 	    my $dir = $self->parse_directive($1);
-	    my $kv = parse_kv($dir->{arg}//"");
-	    if ( $kv && $kv->{toc} ) {
+	    next unless my $kv = parse_kv($dir->{arg}//"");
+	    if ( defined $kv->{toc} ) {
 		$self->{meta}->{_TOC} = [ $kv->{toc} ];
 	    }
+	    if ( $kv->{forceifempty} ) {
+		push( @{ $self->{body} },
+		      { type => "set",
+			name => "forceifempty",
+			value => $kv->{forceifempty} } );
+	    }
 	    next;
 	}
 
@@ -556,43 +567,49 @@ sub parse_song {
 			}
 		    }
 		    $opts = $a->{opts} = { %$opts, %{$a->{opts}} };
+		    unless ( is_true($opts->{omit}) ) {
+			if ( $opts->{align} && $opts->{x} && $opts->{x} =~ /\%$/ ) {
+			    do_warn( "Useless combination of x percentage with align (align ignored)" );
+			    delete $opts->{align};
+	}
 
-		    my $def = !!$id;
-		    $id //= "_Image".$assetid++;
+			my $def = !!$id;
+			$id //= "_Image".$assetid++;
 
-		    if ( defined $opts->{spread} ) {
-			$def++;
-			if ( exists $self->{spreadimage} ) {
-			    do_warn("Skipping superfluous spread image");
+			if ( defined $opts->{spread} ) {
+			    $def++;
+			    if ( exists $self->{spreadimage} ) {
+				do_warn("Skipping superfluous spread image");
+			    }
+			    else {
+				$self->{spreadimage} =
+				  { id => $id, space => $opts->{spread} };
+				warn("Got spread image $id with space=$opts->{spread}\n")
+				  if $config->{debug}->{images};
+			    }
+			}
+
+			# Move to assets.
+			$self->{assets}->{$id} = $a;
+			if ( $def ) {
+			    my $label = delete $a->{label};
+			    do_warn("Label \"$label\" ignored on non-displaying $in_context section\n")
+			      if $label;
 			}
 			else {
-			    $self->{spreadimage} =
-			      { id => $id, space => $opts->{spread} };
-			    warn("Got spread image $id with space=$opts->{spread}\n")
-			      if $config->{debug}->{images};
-			}
-		    }
-
-		    # Move to assets.
-		    $self->{assets}->{$id} = $a;
-		    if ( $def ) {
-			my $label = delete $a->{label};
-			do_warn("Label \"$label\" ignored on non-displaying $in_context section\n")
-			  if $label;
-		    }
-		    else {
-			my $label = delete $opts->{label};
-			$self->add( type => "set",
-				    name => "label",
-				    value => $label )
-			  if $label && $label ne "";
-			$self->add( type => "image",
-				    opts => $opts,
-				    id => $id );
-			if ( $opts->{label} ) {
-			    push( @labels, $opts->{label} )
-			      unless $in_context eq "chorus"
-			      && !$config->{settings}->{choruslabels};
+			    my $label = delete $opts->{label};
+			    $self->add( type => "set",
+					name => "label",
+					value => $label )
+			      if $label && $label ne "";
+			    $self->add( type => "image",
+					opts => $opts,
+					id => $id );
+			    if ( $opts->{label} ) {
+				push( @labels, $opts->{label} )
+				  unless $in_context eq "chorus"
+				  && !$config->{settings}->{choruslabels};
+			    }
 			}
 		    }
 		}
@@ -618,9 +635,9 @@ sub parse_song {
 	if ( /^\s*\{(.*)\}\s*$/ ) {
 	    my $dir = $1;
 	    if ( $prep->{directive} ) {
-		# warn("PRE:  ", $_, "\n");
+		$config->{debug}->{pp} && warn("PRE:  ", $_, "\n");
 		$prep->{directive}->($dir);
-		# warn("POST: ", $_, "\n");
+		$config->{debug}->{pp} && warn("POST: {", $dir, "}\n");
 	    }
 	    $self->add( type => "ignore",
 			text => $_ )
@@ -648,9 +665,9 @@ sub parse_song {
 
 	if ( /\S/ ) {
 	    if ( $prep->{songline} ) {
-		# warn("PRE:  ", $_, "\n");
+		$config->{debug}->{pp} && warn("PRE:  ", $_, "\n");
 		$prep->{songline}->($_);
-		# warn("POST: ", $_, "\n");
+		$config->{debug}->{pp} && warn("POST: ", $_, "\n");
 	    }
 	    if ( $config->{settings}->{flowtext}
 		 && @{ $self->{body}//[] } ) {
@@ -747,14 +764,30 @@ sub parse_song {
     }
 
     if ( $config->{diagrams}->{sorted} ) {
-	@used_chords =
-	  sort ChordPro::Chords::chordcompare @used_chords;
+	sub byname { ChordPro::Chords::chordcompare($a,$b) }
+	@used_chords = sort byname @used_chords;
     }
 
     # For headings, footers, table of contents, ...
     $self->{meta}->{chords} //= [ @used_chords ];
     $self->{meta}->{numchords} = [ scalar(@{$self->{meta}->{chords}}) ];
 
+    if ( %memchords ) {
+	::dump(\%memchords, as => "cc (atend)") if $config->{debug}->{chords};
+    }
+    else {
+	# Avoid clutter.
+    	delete $self->{meta}->{cc};
+    }
+
+    if ( %memchords ) {
+	::dump(\%memchords, as => "cc (atend)") if $config->{debug}->{chords};
+    }
+    else {
+	# Avoid clutter.
+    	delete $self->{meta}->{cc};
+    }
+
     if ( $diagrams =~ /^(user|all)$/ ) {
 	$self->{chords} =
 	  { type   => "diagrams",
@@ -912,7 +945,7 @@ sub decompose {
 	# Normal chords.
 	if ( $chord =~ s/^\[(.*)\]$/$1/ && $chord ne "^" ) {
 	    push(@chords, $chord eq "" ? "" : $self->chord($chord));
-	    if ( $memchords && !$dummy ) {
+	    if ( $memchords && !$dummy && $chord !~ /^\*/ ) {
 		if ( $memcrdinx == 0 ) {
 		    $memorizing++;
 		}
@@ -927,20 +960,20 @@ sub decompose {
 	}
 
 	# Recall memorized chords.
-	elsif ( $memchords && $in_context ) {
+	elsif ( $memchords && $in_context && $chord !~ /^\*/ ) {
 	    if ( $memcrdinx == 0 && @$memchords == 0 ) {
 		do_warn("No chords memorized for $in_context");
-		push( @chords, $chord );
+		push( @chords, $self->chord($chord) );
 		undef $memchords;
 	    }
 	    elsif ( $memcrdinx >= @$memchords ) {
 		do_warn("Not enough chords memorized for $in_context");
-		push( @chords, $chord );
+		push( @chords, $self->chord($chord) );
 	    }
 	    else {
-		push( @chords, $self->chord($memchords->[$memcrdinx]) );
-		warn("Chord recall $in_context\[$memcrdinx]: ", $chords[-1], "\n")
+		warn("Chord recall $in_context\[$memcrdinx]: ", $memchords->[$memcrdinx], "\n")
 		  if $config->{debug}->{chords};
+		push( @chords, $self->chord($memchords->[$memcrdinx]) );
 	    }
 	    $memcrdinx++;
 	}
@@ -968,12 +1001,14 @@ sub cdecompose {
 }
 
 sub decompose_grid {
-    my ($self, $line) = @_;
+    my ($self, $orig) = @_;
+    my $line = fmt_subst( $self, $orig );
+    undef $orig if $orig eq $line;
     $line =~ s/^\s+//;
     $line =~ s/\s+$//;
     return ( tokens => [] ) if $line eq "";
+    local $re_chords = qr/(\[.*?\])/;
 
-    my $orig;
     my %res;
     if ( $line !~ /\|/ ) {
 	$res{margin} = { $self->cdecompose($line), orig => $line };
@@ -1011,9 +1046,15 @@ sub decompose_grid {
 	}
     }
     my $nbt = 0;		# non-bar tokens
+    my $p0;			# this bar chords
+    my $p1;			# prev chords (for % and %% repeat)
+    my $p2;			# pprev chords (for %% repeat)
+    my $si = 0;			# start index
+
     foreach ( @tokens ) {
 	if ( $_ eq "|:" || $_ eq "{" ) {
 	    $_ = { symbol => $_, class => "bar" };
+	    $si = @$memchords if $memchords;
 	}
 	elsif ( /^\|(\d+)(>?)$/ ) {
 	    $_ = { symbol => '|', volta => $1, class => "bar" };
@@ -1021,9 +1062,16 @@ sub decompose_grid {
 	}
 	elsif ( $_ eq ":|" || $_ eq "}" ) {
 	    $_ = { symbol => $_, class => "bar" };
+	    if ( $memchords ) {
+		push( @$memchords, @$memchords[ $si .. $#{$memchords} ] );
+	    }
 	}
 	elsif ( $_ eq ":|:" || $_ eq "}{" ) {
 	    $_ = { symbol => $_, class => "bar" };
+	    if ( $memchords ) {
+		push( @$memchords, @$memchords[ $si .. $#{$memchords} ] );
+		$si = @$memchords;
+	    }
 	}
 	elsif ( $_ eq "|" ) {
 	    $_ = { symbol => $_, class => "bar" };
@@ -1036,9 +1084,26 @@ sub decompose_grid {
 	}
 	elsif ( $_ eq "%" ) {
 	    $_ = { symbol => $_, class => "repeat1" };
+	    if ( $memchords && $p1 ) {
+		push( @$memchords, @$p1 );
+		if ( $config->{debug}->{chords} ) {
+		    warn("Chord memorized for $cctag\[$memcrdinx]: ",
+			 $_, "\n"), $memcrdinx++
+		      for @$p1;
+		}
+	    }
 	}
 	elsif ( $_ eq '%%' ) {
 	    $_ = { symbol => $_, class => "repeat2" };
+	    if ( $memchords && $p1 ) {
+		push( @$memchords, @$p2 ) if $p2;
+		push( @$memchords, @$p1 );
+		if ( $config->{debug}->{chords} ) {
+		    warn("Chord memorized for $cctag\[$memcrdinx]: ",
+			 $_, "\n"), $memcrdinx++
+		      for @$p2, @$p1;
+		}
+	    }
 	}
 	elsif ( $_ eq "/" ) {
 	    $_ = { symbol => $_, class => "slash" };
@@ -1064,8 +1129,20 @@ sub decompose_grid {
 				   : $self->chord($_) } @a ],
 		       class => "chords" };
 	    }
+	    if ( $memchords ) {
+		push( @$memchords, @a );
+		push( @$p0, @a );
+		if ( $config->{debug}->{chords} ) {
+		    warn("Chord memorized for $cctag\[$memcrdinx]: ",
+			 $_, "\n"), $memcrdinx++
+		      for @a;
+		}
+	    }
 	    $nbt++;
 	}
+	if ( $_->{class} eq "bar" ) {
+	    $p2 = $p1; $p1 = $p0; undef $p0;
+	}
     }
     if ( $nbt > $grid_cells->[0] ) {
 	do_warn( "Too few cells for grid content" );
@@ -1122,6 +1199,7 @@ my %abbrevs = (
    cs	      => "chordsize",
    eob	      => "end_of_bridge",
    eoc	      => "end_of_chorus",
+   eog	      => "end_of_grid",
    eot	      => "end_of_tab",
    eov	      => "end_of_verse",
    g	      => "diagrams",
@@ -1131,6 +1209,7 @@ my %abbrevs = (
    ns	      => "new_song",
    sob	      => "start_of_bridge",
    soc	      => "start_of_chorus",
+   sog	      => "start_of_grid",
    sot	      => "start_of_tab",
    sov	      => "start_of_verse",
    st	      => "subtitle",
@@ -1139,6 +1218,10 @@ my %abbrevs = (
    ts         => "textsize",
 	      );
 
+# Use by: runtimeinfo.
+sub _directives { \%directives }
+sub _directive_abbrevs { \%abbrevs }
+
 my $dirpat;
 
 sub parse_directive {
@@ -1152,7 +1235,7 @@ sub parse_directive {
 		     @{$config->{metadata}->{keys}},
 		     keys(%abbrevs),
 		     '(?:start|end)_of_\w+',
-		     '(?:(?:text|chord|chorus|tab|grid|diagrams|title|footer|toc)'.
+		     "(?:$propitems_re".
 		     '(?:font|size|colou?r))',
 		) . ')';
 	$dirpat = qr/$dirpat/;
@@ -1216,12 +1299,19 @@ sub directive {
     my $dd = $self->parse_directive($d);
     return 1 if $dd->{omit} == 1;
 
+    my $dir = $dd->{name};
     my $arg = $dd->{arg};
     if ( $arg ne "" ) {
 	$arg = fmt_subst( $self, $arg );
-	return 1 if $arg !~ /\S/;
+	if ( $arg !~ /\S/ ) { 	# expansion yields empty
+	    if ( $dir =~ /^start_of_/ ) {
+		$dd->{omit} = 2;
+	    }
+	    else {
+		return 1;
+	    }
+	}
     }
-    my $dir = $dd->{name};
 
     if ( $directives{$dir} ) {
 	return $directives{$dir}->( $self, $dir, $arg, $dd->{arg} );
@@ -1240,18 +1330,23 @@ sub directive {
 	}
 	@chorus = (), $chorus_xpose = $chorus_xpose_dir = 0
 	  if $in_context eq "chorus";
+	undef $cctag;
+
 	if ( $in_context eq "grid" ) {
-	    if ( $arg eq "" ) {
+	    $cctag = "grid";
+	    my $kv = parse_kv( $arg, "shape" );
+	    my $shape = $kv->{shape} // "";
+	    if ( $shape eq "" ) {
 		$self->add( type => "set",
 			    name => "gridparams",
 			    value => $grid_arg );
 	    }
-	    elsif ( $arg =~ m/^
+	    elsif ( $shape =~ m/^
 			      (?: (\d+) \+)?
 			      (\d+) (?: x (\d+) )?
 			      (?:\+ (\d+) )?
 			      (?:[:\s+] (.*)? )? $/x ) {
-		do_warn("Invalid grid params: $arg (must be non-zero)"), return
+		do_warn("Invalid grid params: $shape (must be non-zero)"), return
 		  unless $2;
 		$grid_arg = [ $2, $3//1, $1//0, $4//0 ];
 		$self->add( type => "set",
@@ -1259,30 +1354,39 @@ sub directive {
 			    value =>  [ @$grid_arg, $5||"" ] );
 		push( @labels, $5 ) if length($5||"");
 	    }
-	    elsif ( $arg ne "" ) {
+	    elsif ( $shape ne "" ) {
 		$self->add( type => "set",
 			    name => "gridparams",
-			    value =>  [ @$grid_arg, $arg ] );
-		push( @labels, $arg );
+			    value =>  [ @$grid_arg, $shape ] );
+		push( @labels, $shape );
+	    }
+	    if ( ($kv->{label}//"") ne "" ) {
+		$self->add( type  => "set",
+			    name  => "label",
+			    value => $kv->{label} );
+		push( @labels, $kv->{label} );
+	    }
+
+	    # Grid sections always memorize unless "cc=".
+	    if ( ($kv->{cc}//="grid") ne "" ) {
+		$cctag = $kv->{cc};
+		$memchords = $memchords{$cctag} //= [];
+		$memcrdinx = 0;
+		$memorizing = 1;
 	    }
 	    $grid_cells = [ $grid_arg->[0] * $grid_arg->[1],
 			    $grid_arg->[2],  $grid_arg->[3] ];
+	    return 1;
 	}
 	elsif ( exists $config->{delegates}->{$in_context} ) {
 	    my $d = $config->{delegates}->{$in_context};
-	    my $label = $arg;
 	    my %opts;
 	    if ( $xpose || $config->{settings}->{transpose} ) {
 		$opts{transpose} =
 		  $xpose + ($config->{settings}->{transpose}//0 );
 	    }
-	    my $kv = {};
-	    if ( $arg =~ /\b(id|label|scale|split|spread|width|align|center)=(.+)/ ) {
-		$kv = parse_kv($arg);
-	    }
-	    else {
-		$kv->{label} = $arg if $arg ne "";
-	    }
+	    my $kv = parse_kv( $arg, "label" );
+	    delete $kv->{label} if ($kv->{label}//"") eq "";
 	    $self->add( type     => "image",
 			subtype  => "delegate",
 			delegate => $d->{module},
@@ -1291,32 +1395,72 @@ sub directive {
 			opts     => { %opts, %$kv },
 			exists($kv->{id}) ? ( id => $kv->{id} ) : (),
 			open     => 1 );
-	    push( @labels, $kv->{label} ) if $kv->{label}//"" ne "";
+	    push( @labels, $kv->{label} ) if exists $kv->{label};
 	}
 	elsif ( $arg ne "" ) {
-	    # Prefer explicit label.
-	    if ( $arg =~ /^label=/ ) {
-		$arg = parse_kv($arg)->{label};
-	    }
-	    elsif ( $arg =~ /\b(id|scale|split|spread|width|align|center)=(.+)/ ) {
-		# Doesn't look like a label. Assume a mistake.
+	    my $kv = parse_kv( $arg, "label" );
+	    my $label = delete $kv->{label};
+	    my $chords = delete $kv->{cc};
+	    if ( %$kv ) {
+		# Assume a mistake.
 		do_warn("Garbage in start_of_$in_context: $arg (ignored)\n");
-		$arg = "";
 	    }
-	    if ( $arg ne "" ) {
+	    elsif ( $label ) {
 		$self->add( type  => "set",
 			    name  => "label",
-			    value => $arg );
-		push( @labels, $arg )
+			    value => $label );
+		push( @labels, $label)
 		  unless $in_context eq "chorus"
 		  && !$config->{settings}->{choruslabels};
 	    }
+	    if ( $chords ) {
+		$chords =~ s/^\s*(.*)\s*/$1/;
+		$cctag = $in_context;
+		# Do we have a name? Chords? Both?
+		# name:C D E
+		# :C D E
+		# :
+		if ( $chords =~ /^(\w*):(.*)/ ) {
+		    # Name, possibly empty.
+		    $cctag = $1 if length($1);
+		    # Chords, possibly empty.
+		    $chords = $2;
+		}
+		# C D E
+		elsif ( $chords =~ /\s/ ) {
+		    # Whitespace separated chords.
+		}
+		# name
+		elsif ( $chords =~ /^\w+$/ ) {
+		    $cctag = $chords;
+		    $chords = "";
+		}
+		# ???
+		else {
+		    warn("Unrecognized cc value: \"$chords\"\n")
+		      if $chords;
+		    $chords = "";
+		}
+		if ( $chords ne "" ) {
+		    $memchords = [ split( ' ', $chords ) ];
+		    $memchords{$cctag} = $memchords;
+		    $memcrdinx = 0;
+		    $memorizing = 0;
+		    if ( $config->{debug}->{chords} ) {
+			my $i = 0;
+			warn("Chord memorized for $cctag\[$i]: ",
+			     $_, "\n"), $i++
+			       for @$memchords;
+		    }
+		    return 1;
+		}
+	    }
 	}
 
 	# Enabling this always would allow [^] to recall anyway.
 	# Feature?
 	if ( $config->{settings}->{memorize} ) {
-	    $memchords = $memchords{$in_context} //= [];
+	    $memchords = ($memchords{$cctag//$in_context} //= []);
 	    $memcrdinx = 0;
 	    $memorizing = 0;
 	}
@@ -1341,8 +1485,7 @@ sub directive {
     }
 
     # Formatting. {chordsize XX} and such.
-    if ( $dir =~ m/ ^( text | chord | chorus | tab | grid | diagrams
-		       | title | footer | toc )
+    if ( $dir =~ m/ ^( $propitems_re )
 		     ( font | size | colou?r )
 		     $/x ) {
 	my $item = $1;
@@ -1374,49 +1517,10 @@ sub directive {
 		    value => $arg,
 		  );
 
-	# THIS IS BASICALLY A COPY OF THE CODE IN Config.pm.
-	# TODO: GENERALIZE.
-	my $ccfg = {};
-	my @k = split( /[:.]/, $1 );
-	my $c = \$ccfg;		# new
-	my $o = $config;	# current
-	my $lk = pop(@k);	# last key
+	$config->unlock;
+	prpadd2cfg( $config, $1 => $arg );
+	$config->lock;
 
-	# Step through the keys.
-	foreach ( @k ) {
-	    $c = \($$c->{$_});
-	    $o = $o->{$_};
-	}
-
-	# Turn hash.array into hash.array.> (append).
-	if ( ref($o) eq 'HASH' && ref($o->{$lk}) eq 'ARRAY' ) {
-	    $c = \($$c->{$lk});
-	    $o = $o->{$lk};
-	    $lk = '>';
-	}
-
-	# Final key. Merge array if so.
-	if ( ( $lk =~ /^\d+$/ || $lk eq '>' || $lk eq '<' )
-	       && ref($o) eq 'ARRAY' ) {
-	    unless ( ref($$c) eq 'ARRAY' ) {
-		# Only copy orig values the first time.
-		$$c->[$_] = $o->[$_] for 0..scalar(@{$o})-1;
-	    }
-	    if ( $lk eq '>' ) {
-		push( @{$$c}, $arg );
-	    }
-	    elsif ( $lk eq '<' ) {
-		unshift( @{$$c}, $arg );
-	    }
-	    else {
-		$$c->[$lk] = $arg;
-	    }
-	}
-	else {
-	    $$c->{$lk} = $arg;
-	}
-
-	$config->augment($ccfg);
 	upd_config();
 
 	return 1;
@@ -1440,18 +1544,20 @@ sub dir_chorus {
     my $chorus = @chorus ? dclone(\@chorus) : [];
 
     if ( @$chorus && $arg && $arg ne "" ) {
+	my $kv = parse_kv( $arg, "label" );
+	my $label = $kv->{label};
 	if ( $chorus->[0]->{type} eq "set" && $chorus->[0]->{name} eq "label" ) {
-	    $chorus->[0]->{value} = $arg;
+	    $chorus->[0]->{value} = $label;
 	}
-	else {
+	elsif ( defined $label ) {
 	    unshift( @$chorus,
 		     { type => "set",
 		       name => "label",
-		       value => $arg,
+		       value => $label,
 		       context => "chorus",
 		     } );
 	}
-	push( @labels, $arg )
+	push( @labels, $label )
 	  if $config->{settings}->{choruslabels};
     }
 
@@ -1521,7 +1627,18 @@ sub dir_comment {
 sub dir_image {
     my ( $self, $dir, $arg ) = @_;
     return 1 if $::running_under_test && !$arg;
-    my $res = parse_kv($arg);
+    use Text::ParseWords qw(quotewords);
+    my @words = quotewords( '\s+', 1, $arg );
+    my $res;
+    # Imply src= if word 0 is not kv.
+    if ( @words && $words[0] !~ /\w+=/ ) {
+	$words[0] = "src=" . $words[0];
+	$res = parse_kv( \@words );
+    }
+    else {
+	$res = parse_kv( \@words, "src" );
+    }
+
     my $uri;
     my $id;
     my $chord;
@@ -1531,7 +1648,7 @@ sub dir_image {
 	if ( $k =~ /^(title)$/i && $v ne "" ) {
 	    $opts{lc($k)} = $v;
 	}
-	elsif ( $k =~ /^(border|spread|center|persist)$/i
+	elsif ( $k =~ /^(border|spread|center|persist|omit)$/i
 		&& $v =~ /^(\d+)$/ ) {
 	    if ( $k eq "center" && $v ) {
 		$opts{align} = $k;
@@ -1545,14 +1662,15 @@ sub dir_image {
 	    $opts{lc($k)} = $v;
 	}
 	elsif ( $k =~ /^(x|y)$/i
-		&& $v =~ /^([-+]?\d+(?:\.\d+)?\%?)$/ ) {
+		&& $v =~ /^(?:base[+-])?([-+]?\d+(?:\.\d+)?\%?)$/ ) {
 	    $opts{lc($k)} = $v;
 	}
 	elsif ( $k =~ /^(scale)$/
-		&& $v =~ /^(\d+(?:\.\d+)?)(%)?$/ ) {
-	    $opts{lc($k)} = $2 ? $1/100 : $1;
+		&& $v =~ /^(\d+(?:\.\d+)?)(%)?(?:,(\d+(?:\.\d+)?)(%)?)?$/ ) {
+	    $opts{lc($k)} = [ $2 ? $1/100 : $1 ];
+	    $opts{lc($k)}->[1] = $3 ? $4 ? $3/100 : $3 : $opts{lc($k)}->[0];
 	}
-	elsif ( $k =~ /^(center|border|spread|persist)$/i ) {
+	elsif ( $k =~ /^(center|border|spread|persist|omit)$/i ) {
 	    if ( $k eq "center" ) {
 		$opts{align} = $k;
 	    }
@@ -1572,11 +1690,11 @@ sub dir_image {
 	elsif ( $k =~ /^(type)$/i && $v ne "" ) {
 	    $opts{type} = $v;
 	}
-	elsif ( $k =~ /^(label)$/i && $v ne "" ) {
+	elsif ( $k =~ /^(label|href)$/i && $v ne "" ) {
 	    $opts{lc($k)} = $v;
 	}
 	elsif ( $k =~ /^(anchor)$/i
-		&& $v =~ /^(paper|page|column|float|line)$/ ) {
+		&& $v =~ /^(paper|page|allpages|column|float|line)$/ ) {
 	    $opts{lc($k)} = lc($v);
 	}
 	elsif ( $k =~ /^(align)$/i
@@ -1593,16 +1711,23 @@ sub dir_image {
 	}
     }
 
+    return if is_true($opts{omit});
+
     unless ( $uri || $id || $chord ) {
 	do_warn( "Missing image source\n" );
 	return;
     }
+    if ( $opts{align} && $opts{x} && $opts{x} =~ /\%$/ ) {
+	do_warn( "Useless combination of x percentage with align (align ignored)" );
+	delete $opts{align};
+    }
 
     # If the image uri does not have a directory, look it up
     # next to the song, and then in the images folder of the
     # resources.
     if ( $uri && CP->is_here($uri) ) {
-	my $found = CP->siblingres( $diag->{file}, $uri, class => "images" );
+	my $found = CP->siblingres( $diag->{file}, $uri, class => "images" )
+	  || CP->siblingres( $diag->{file}, $uri, class => "icons" );
 	if ( $found ) {
 	    $uri = $found;
 	}
@@ -1630,12 +1755,17 @@ sub dir_image {
     # Store as asset.
     if ( $uri ) {
 	my $opts;
-	$opts->{type} = $opts{type}    if $opts{type};
-	$opts->{persist} = $opts{persist} if $opts{persist};
-	delete $opts{$_} for qw( type persist );
+	for ( qw( type persist href ) ) {
+	    $opts->{$_} = $opts{$_} if defined $opts{$_};
+	    delete $opts{$_};
+	}
+	for ( qw( spread ) ) {
+	    $opts->{$_} = $opts{$_} if defined $opts{$_};
+	}
 
 	if ( $id && %opts ) {
-	    do_warn("Asset definition \"$id\" does not take attributes");
+	    do_warn("Asset definition \"$id\" does not take attributes",
+		   " (" . join(" ",sort keys %opts) . ")");
 	    return;
 	}
 
@@ -1708,6 +1838,9 @@ sub dir_meta {
 	    @vals = map { s/s\+$//; $_ }
 	      split( quotemeta($config->{metadata}->{separator}), $vals[0] );
 	}
+	else {
+	    pop(@vals) if $vals[0] eq '';
+	}
 	my $m = $self->{meta};
 
 	# User and instrument cannot be set here.
@@ -1727,8 +1860,11 @@ sub dir_meta {
 		    local( $self->{chordsinfo}->{_dummy_} ) = { root_ord => 0 };
 		    $self->parse_chord($val);
 		};
+		do_warn("Illegal key: \"$val\"\n"), next unless $info;
 		my $name = $info->name;
 		my $act = $name;
+		$info->{key} = $name
+		  unless $config->{settings}->{'enharmonic-transpose'};
 
 		if ( $capo ) {
 		    $act = $self->add_chord( $info->transpose($capo) );
@@ -2065,7 +2201,7 @@ sub define_chord {
 	# frets N N ... N
 	elsif ( $a eq "frets" ) {
 	    my @f;
-	    while ( @a && $a[0] =~ /^(?:[0-9]+|[-xXN])$/ && @f < $strings ) {
+	    while ( @a && $a[0] =~ /^(?:-?[0-9]+|[-xXN])$/ && @f < $strings ) {
 		push( @f, shift(@a) );
 	    }
 	    if ( @f == $strings ) {
@@ -2312,6 +2448,8 @@ sub parse_chord {
     if ( $info ) {
 	warn( "Parsing chord: \"$chord\" found \"",
 	      $info->name, "\" in ", $info->{_via}, "\n" ) if $debug > 1;
+	return ChordPro::Chord::NC->new( { name => $info->name } )
+	  if $info->is_nc;
 	$info->dump if $debug > 1;
     }
     else {
@@ -2348,7 +2486,9 @@ sub parse_chord {
 	$info = $i;
 	warn( "Parsing chord: \"$chord\" transposed ",
 	      sprintf("%+d", $xp), " to \"",
-	      $info->name, "\"\n" ) if $debug > 1;
+	      $info->name, "\"",
+	      ( $self->{meta}->{key} ? (" key ".$self->{meta}->{key}->[-1]) : ()),
+	      "\n" ) if $debug > 1;
     }
     # else: warning has been given.
 
@@ -2394,6 +2534,7 @@ sub parse_chord {
 	warn( "Parsing chord: \"$chord\" transcoded to ",
 	      $info->name,
 	      " (", $info->{system}, ")",
+	      defined($key_ord) ? " key ".$self->{meta}->{key}->[-1] : "",
 	      "\n" ) if $debug > 1;
 	if ( my $i = ChordPro::Chords::known_chord($info) ) {
 	    warn( "Parsing chord: \"$chord\" found \"",
@@ -2422,6 +2563,8 @@ sub parse_chord {
     }
 
     if ( $info ) {
+	$info->{key} = $self->{meta}->{key}->[-1]
+	  unless $config->{settings}->{'enharmonic-transpose'};
 	warn( "Parsing chord: \"$chord\" okay: \"",
 	      $info->name, "\" \"",
 	      $info->chord_display, "\"",
@@ -2488,23 +2631,4 @@ sub dump {
     ::dump($a);
 }
 
-unless ( caller ) {
-    require DDumper;
-    binmode STDERR => ':utf8';
-    ChordPro::Config::configurator();
-    my $s = ChordPro::Song->new;
-    $options->{settings}->{transpose} = 0;
-    for ( @ARGV ) {
-	if ( /^[a-z]/ ) {
-	    $options->{settings}->{transcode} = $_;
-	    next;
-	}
-#	DDumper::DDumper( $s->parse_chord($_) );
-	my ( undef, $i ) = $s->parse_chord($_);
-	warn("$_ => ", $i->name, " => ", $s->add_chord($i, $i->name eq 'D'), "\n" );
-	$xpose++;
-    }
-    DDumper::DDumper($s->{chordsinfo});
-}
-
 1;
